#include "stdafx.h"

#include "../System.h"
#include "../MemMap.h"
#include "../Misc.h"
#include "../Host.h"

#include "PSPDisplay.h"
#include "PSPGfxState.h"

#include "PSPSystem.h"
#include "ge_constants.h"

#include "PSPDisplayListInterpreter.h"

inline void glEnDis(GLuint cmd, int value)
{
	(value ? glEnable : glDisable)(cmd);
}

u32 pc;
u32 stallAddr;

extern u32 curTextureWidth;
extern u32 curTextureHeight;

int dlIdGenerator=1;

struct DisplayList
{
	int id;
	u32 listpc;
	u32 stall;
};

std::vector<DisplayList> dlQueue;

bool ProcessDLQueue()
{
	bool doMore = false;
	std::vector<DisplayList>::iterator iter = dlQueue.begin();
	while (!(iter == dlQueue.end()))
	{
		DisplayList &l = *iter;
		pc = l.listpc;
		stallAddr = l.stall;
//		LOG(G3D,"Okay, starting DL execution at %08 - stall = %08x", pc, stallAddr);
		if (!PSPDisplay_InterpretList())
		{
			l.listpc = pc;
			l.stall = stallAddr;
			return false;
		}
		else
		{
			//At the end, we can remove it from the queue and continue
			dlQueue.erase(iter);
			//this invalidated the iterator, let's fix it
			iter = dlQueue.begin();
		}
	}
	return true; //no more lists!
}

u32 PSPDisplay_EnqueueList(u32 listpc, u32 stall)
{
	DisplayList dl;
	dl.id = dlIdGenerator++;
	dl.listpc = listpc&0xFFFFFFF;
	dl.stall = stall&0xFFFFFFF;
	dlQueue.push_back(dl);
	if (!ProcessDLQueue())
		return dl.id;
	else
		return 0;
}

void PSPDisplay_UpdateStall(int listid, u32 newstall)
{
	for (std::vector<DisplayList>::iterator iter = dlQueue.begin(); iter!=dlQueue.end(); iter++)
	{
		DisplayList &l = *iter;
		if (l.id == listid)
		{
			l.stall = newstall & 0xFFFFFFF;
		}
	}
	
	ProcessDLQueue();
}


//Put on second core?
	
bool PSPDisplay_InterpretList()
{
	u32 stack[2];
	u32 stackptr = 0;
	int count=0;
	u32 prev = 0;
	u32 op = 0;
	while (true)
	{
		//if (count++ > 1000) return false;
		if (pc == stallAddr)
			return false;
		//LOG(G3D,"pc: %08x", pc);
		prev = op;

		op = ReadMem32Unchecked(pc); //read from memory
		u32 data = op&0xFFFFFF;
		gstate.cmdmem[op>>24] = data;		

		u32 oldpc=pc;
		switch (op>>24) 
		{
		case GE_CMD_BASE:
			break;

		case GE_CMD_VADDR: 
			gstate.vertexAddr = (gstate.base<<8)|data;
			LOG(G3D,"DL VADDR: %06x", gstate.vertexAddr);
			break;

		case GE_CMD_IADDR:  
			gstate.indexAddr  = (gstate.base<<8)|data;
			LOG(G3D,"DL IADDR: %06x", gstate.indexAddr);
			break;

		case GE_CMD_PRIM:
			{
				u32 count = data & 0xFFFF;
				u32 type = data >> 16;
				static const char* types[7] = {
					"POINTS=0,",
					"LINES=1,",
					"LINE_STRIP=2,",
					"TRIANGLES=3,",
					"TRIANGLE_STRIP=4,",
					"TRIANGLE_FAN=5,",
					"RECTANGLES=6,",
				};
				LOG(G3D, "DrawPrim type: %s  count: %i", type<7 ? types[type] : "INVALID", count); 
				LOG(G3D, "DrawPrim vaddr= %08x, iaddr= %08x", gstate.vertexAddr, gstate.indexAddr);
				TransformAndDrawPrim(GetMemPointer(gstate.vertexAddr), GetMemPointer(gstate.indexAddr), type, count);
			}
			break;

		case GE_CMD_JUMP: 
			{
				u32 target = ((gstate.base<<8) | (op & 0xFFFFFC)) & 0x0FFFFFFF;
				LOG(G3D,"DL CMD JUMP - %08x to %08x", pc, target);
				pc = target;
			}
			continue; //do not increase pc

		case GE_CMD_CALL: 
			{
				u32 retval = pc+4;
				stack[stackptr++] = retval; 
				u32 target = ((gstate.base<<8) | (op & 0xFFFFFC)) & 0xFFFFFFF;
				LOG(G3D,"DL CMD CALL - %08x to %08x, ret=%08x", pc, target, retval);
				pc = target;
			}
			continue;//do not increase pc

		case GE_CMD_RET: 
			//TODO : debug!
			{
				u32 target = stack[--stackptr] & 0xFFFFFFF; 
				LOG(G3D,"DL CMD RET - from %08x to %08x", pc, target);
				pc = target;
			}
			continue;//do not increase pc

		case GE_CMD_ORIGIN:
			gstate.offsetAddr = pc & 0xFFFFFF;
			break;

		case GE_CMD_VERTEXTYPE:
			LOG(G3D,"DL SetVertexType: %06x", data);
			break;

		case GE_CMD_OFFSETADDR:
			//			offsetAddr = data<<8;
			break;

		case GE_CMD_SIGNAL:
			break;

		case GE_CMD_FINISH:
			LOG(G3D,"DL CMD FINISH");
			break;

		case GE_CMD_END: 
			LOG(G3D,"DL CMD END");
			{
				switch (prev >> 24)
				{
				case GE_CMD_FINISH:
					return true;
				default:
					LOG(G3D,"Ah, not finished: %06x", prev & 0xFFFFFF);
					break;
				}
			}
			break;

		case GE_CMD_REGION1:
			{
				int x1 = data & 0x3ff;
				int y1 = data >> 10;
				//topleft
				LOG(G3D,"DL Region TL: %d %d", x1, y1);
			}
			break;

		case GE_CMD_REGION2:
			{
				int x2 = data & 0x3ff;
				int y2 = data >> 10;
				LOG(G3D,"DL Region BR: %d %d", x2, y2);
			}
			break;

		case GE_CMD_CLIPENABLE:
			//we always clip, this is opengl
			break;

		case GE_CMD_CULLFACEENABLE: 
			glEnDis(GL_CULL_FACE, data); 
			break;

		case GE_CMD_TEXTUREMAPENABLE: 
			LOG(G3D, "Texture map enable: %i", data);
			glEnDis(GL_TEXTURE_2D, data&1); 
			break;

		case GE_CMD_LIGHTINGENABLE:
			LOG(G3D, "Lighting enable: %i", data);
			//We don't use OpenGL lighting
			break;

		case GE_CMD_FOGENABLE:		
			LOG(G3D, "Fog Enable: %i", gstate.fogEnable);
			break;

		case GE_CMD_DITHERENABLE:
			LOG(G3D, "Dither Enable: %i", gstate.ditherEnable);
			break;

		case GE_CMD_OFFSETX:		
			LOG(G3D, "Offset X: %i", gstate.offsetx);
			break;

		case GE_CMD_OFFSETY:		
			LOG(G3D, "Offset Y: %i", gstate.offsety);
			break;

		case GE_CMD_TEXSCALEU: 
			gstate.uScale = getFloat24(data); 
			LOG(G3D, "Texture U Scale: %f", gstate.uScale);
			break;

		case GE_CMD_TEXSCALEV: 
			gstate.vScale = getFloat24(data); 
			LOG(G3D, "Texture V Scale: %f", gstate.vScale);
			break;

		case GE_CMD_TEXOFFSETU: 
			gstate.uOff = getFloat24(data);  
			LOG(G3D, "Texture U Offset: %f", gstate.uOff);
			break;

		case GE_CMD_TEXOFFSETV: 
			gstate.vOff = getFloat24(data);  
			LOG(G3D, "Texture V Offset: %f", gstate.vOff);
			break;

		case GE_CMD_SCISSOR1:
			{
				int x1 = data & 0x3ff;
				int y1 = data >> 10;
				LOG(G3D, "Scissor TL: %i, %i", x1,y1);
			}
			break;
		case GE_CMD_SCISSOR2:
			{
				int x2 = data & 0x3ff;
				int y2 = data >> 10;
				LOG(G3D, "Scissor BR: %i, %i", x2, y2);
			}
			break;

		case GE_CMD_MINZ: 
			break;

		case GE_CMD_MAXZ: 
			break;

		case GE_CMD_FRAMEBUFPTR:
			{
				u32 ptr = op & 0xFFE000;
				LOG(G3D, "DL FramebufPtr: %08x", gstate.fbptr);
			}
			break;

		case GE_CMD_FRAMEBUFWIDTH:
			{
				u32 w = data & 0xFFFFFF;
				LOG(G3D, "DL FramebufWidth: %i", gstate.fbwidth);
			}
			break;

		case GE_CMD_FRAMEBUFPIXFORMAT:
			break;

		case GE_CMD_TEXADDR0:
			gstate.textureChanged=true;
		case GE_CMD_TEXADDR1:
		case GE_CMD_TEXADDR2:
		case GE_CMD_TEXADDR3:
		case GE_CMD_TEXADDR4:
		case GE_CMD_TEXADDR5:
		case GE_CMD_TEXADDR6:
		case GE_CMD_TEXADDR7:
			LOG(G3D,"DL Texture address %i: %06x", (op>>24)-GE_CMD_TEXADDR0, data);
			break;

		case GE_CMD_TEXBUFWIDTH0:
			gstate.textureChanged=true;
		case GE_CMD_TEXBUFWIDTH1:
		case GE_CMD_TEXBUFWIDTH2:
		case GE_CMD_TEXBUFWIDTH3:
		case GE_CMD_TEXBUFWIDTH4:
		case GE_CMD_TEXBUFWIDTH5:
		case GE_CMD_TEXBUFWIDTH6:
		case GE_CMD_TEXBUFWIDTH7:
			LOG(G3D,"DL Texture BUFWIDTHess %i: %06x", (op>>24)-GE_CMD_TEXBUFWIDTH0, data);
			break;

		case GE_CMD_CLUTADDR:
			//LOG(G3D,"CLUT base addr: %06x", data);
			break;

		case GE_CMD_CLUTADDRUPPER:
			LOG(G3D,"CLUT addr: %08x", ((gstate.clutaddrupper&0xFF0000)<<8) | gstate.clutaddr);
			break;

		case GE_CMD_LOADCLUT:
			{
				u32 clutAttr = ((gstate.clutaddrupper&0xFF0000)<<8) | gstate.clutaddr;
				u16 *clut = (u16*)GetMemPointer(clutAttr);
				int numColors = 16*(data&0x3F);
				memcpy(gstate.paletteMem, clut, numColors*2);
				LOG(G3D,"Clut load: %i palettes", data);
			}
			break;

//		case GE_CMD_TRANSFERSRC:

		case GE_CMD_TRANSFERSRCW:
			{
				u32 xferSrc = gstate.transfersrc | ((data&0xFF0000)<<8);
				u32 xferSrcW = gstate.transfersrcw & 1023;
				LOG(G3D,"Block Transfer Src: %08x  W: %i", xferSrc, xferSrcW);
				break;
			}
//		case GE_CMD_TRANSFERDST:

		case GE_CMD_TRANSFERDSTW:
			{
				u32 xferDst= gstate.transferdst | ((data&0xFF0000)<<8);
				u32 xferDstW = gstate.transferdstw & 1023;
				LOG(G3D,"Block Transfer Dest: %08x  W: %i", xferDst, xferDstW);
				break;
			}
		
		case GE_CMD_TRANSFERSRCPOS:
			{
				u32 x = (data & 1023)+1;
				u32 y = ((data>>10) & 1023)+1;
				LOG(G3D,"Block Transfer Src Rect TL: %i, %i", x, y);
				break;
			}

		case GE_CMD_TRANSFERDSTPOS:
			{
				u32 x = (data & 1023)+1;
				u32 y = ((data>>10) & 1023)+1;
				LOG(G3D,"Block Transfer Dest Rect TL: %i, %i", x, y);
				break;
			}

		case GE_CMD_TRANSFERSIZE:
			{
				u32 w = (data & 1023)+1;
				u32 h = ((data>>10) & 1023)+1;
				LOG(G3D,"Block Transfer Rect Size: %i x %i", w, h);
				break;
			}

		case GE_CMD_TRANSFERSTART:
			{
				LOG(G3D,"Texture Transfer Start: PixFormat %i", data);
				//MessageBox(0,"Block Xfer",0,0);
				break;
			}

		case GE_CMD_TEXSIZE0:
			gstate.textureChanged=true;
			{
				int w = 1 << (gstate.texsize[0] & 0xf);
				int h = 1 << ((gstate.texsize[0]>>8) & 0xf);
				curTextureHeight=h;
				curTextureWidth=w;
			}
			//fall thru
		case GE_CMD_TEXSIZE1:
		case GE_CMD_TEXSIZE2:
		case GE_CMD_TEXSIZE3:
		case GE_CMD_TEXSIZE4:
		case GE_CMD_TEXSIZE5:
		case GE_CMD_TEXSIZE6:
		case GE_CMD_TEXSIZE7:
			LOG(G3D,"DL Texture Size: %06x",  data);
			break;
		case GE_CMD_ZBUFPTR:
			{
				u32 ptr = op&0xFFE000;
				LOG(G3D,"Zbuf Ptr: %06x",  data);
			}
			break;
		case GE_CMD_ZBUFWIDTH:
			{
				u32 w = data & 0xFFFFFF;
				LOG(G3D,"Zbuf Width: %i",  data);
			}
			break;

		case GE_CMD_AMBIENTCOLOR:
			LOG(G3D,"DL Ambient Color: %06x",  data);
			break;

		case GE_CMD_AMBIENTALPHA:
			LOG(G3D,"DL Ambient Alpha: %06x",  data);
			break;

		case GE_CMD_MATERIALAMBIENT:
			LOG(G3D,"DL Material Ambient Color: %06x",  data);
			break;

		case GE_CMD_MATERIALDIFFUSE:
			LOG(G3D,"DL Material Diffuse Color: %06x",  data);
			break;

		case GE_CMD_MATERIALEMISSIVE:
			LOG(G3D,"DL Material Emissive Color: %06x",  data);
			break;

		case GE_CMD_MATERIALSPECULAR:
			LOG(G3D,"DL Material Specular Color: %06x",  data);
			break;

		case GE_CMD_MATERIALALPHA:
			LOG(G3D,"DL Material Alpha Color: %06x",  data);
			break;

		case GE_CMD_MATERIALSPECULARCOEF:
			LOG(G3D,"DL Material specular coef: %f", getFloat24(data));
			break;

		case GE_CMD_LIGHTTYPE0:
		case GE_CMD_LIGHTTYPE1:
		case GE_CMD_LIGHTTYPE2:
		case GE_CMD_LIGHTTYPE3:
			LOG(G3D,"DL Light %i type: %06x", (op>>24)-GE_CMD_LIGHTTYPE0, data);
			break;

		case GE_CMD_LX0:case GE_CMD_LY0:case GE_CMD_LZ0:
		case GE_CMD_LX1:case GE_CMD_LY1:case GE_CMD_LZ1:
		case GE_CMD_LX2:case GE_CMD_LY2:case GE_CMD_LZ2:
		case GE_CMD_LX3:case GE_CMD_LY3:case GE_CMD_LZ3:
			{
				int n = (op>>24) - GE_CMD_LX0;
				int l = n / 3;
				int c = n % 3;
				float val = getFloat24(data);
				LOG(G3D,"DL Light %i %c pos: %f", l, c+'X', val);
				gstate.lightpos[l][c] = val;
			}
			break;

		case GE_CMD_LDX0:case GE_CMD_LDY0:case GE_CMD_LDZ0:
		case GE_CMD_LDX1:case GE_CMD_LDY1:case GE_CMD_LDZ1:
		case GE_CMD_LDX2:case GE_CMD_LDY2:case GE_CMD_LDZ2:
		case GE_CMD_LDX3:case GE_CMD_LDY3:case GE_CMD_LDZ3:
			{
				int n = (op>>24) - GE_CMD_LDX0;
				int l = n / 3;
				int c = n % 3;
				float val = getFloat24(data);
				LOG(G3D,"DL Light %i %c dir: %f", l, c+'X', val);
				gstate.lightdir[l][c] = val;
			}
			break;

		case GE_CMD_LKA0:case GE_CMD_LKB0:case GE_CMD_LKC0:
		case GE_CMD_LKA1:case GE_CMD_LKB1:case GE_CMD_LKC1:
		case GE_CMD_LKA2:case GE_CMD_LKB2:case GE_CMD_LKC2:
		case GE_CMD_LKA3:case GE_CMD_LKB3:case GE_CMD_LKC3:
			{
				int n = (op>>24) - GE_CMD_LKA0;
				int l = n / 3;
				int c = n % 3;
				float val = getFloat24(data);
				LOG(G3D,"DL Light %i %c att: %f", l, c+'X', val);
				gstate.lightatt[l][c] = val;
			}
			break;


		case GE_CMD_LAC0:case GE_CMD_LAC1:case GE_CMD_LAC2:case GE_CMD_LAC3:
		case GE_CMD_LDC0:case GE_CMD_LDC1:case GE_CMD_LDC2:case GE_CMD_LDC3:
		case GE_CMD_LSC0:case GE_CMD_LSC1:case GE_CMD_LSC2:case GE_CMD_LSC3:
			{
				float r = (float)(data>>16)/255.0f;
				float g = (float)((data>>8) & 0xff)/255.0f;
				float b = (float)(data & 0xff)/255.0f;

				int l = ((op>>24) - GE_CMD_LAC0)/3;
				int t = ((op>>24) - GE_CMD_LAC0)%3;
				gstate.lightColor[t][l].r = r;
				gstate.lightColor[t][l].g = g;
				gstate.lightColor[t][l].b = b;
			}
			break;

		case GE_CMD_VIEWPORTX1:
		case GE_CMD_VIEWPORTY1:
		case GE_CMD_VIEWPORTZ1:
		case GE_CMD_VIEWPORTX2:
		case GE_CMD_VIEWPORTY2:
		case GE_CMD_VIEWPORTZ2:
			LOG(G3D,"DL Viewport param %i: %f", (op>>24)-GE_CMD_VIEWPORTX1, getFloat24(data));
			break;
		case GE_CMD_LIGHTENABLE0:
		case GE_CMD_LIGHTENABLE1:
		case GE_CMD_LIGHTENABLE2:
		case GE_CMD_LIGHTENABLE3:
			LOG(G3D,"DL Light %i enable: %d", (op>>24)-GE_CMD_LIGHTENABLE0, data);
			break;
		case GE_CMD_CULL:
			LOG(G3D,"DL cull: %06x", data);
			glCullFace(data ? GL_BACK : GL_FRONT);
			break;

		case GE_CMD_SHADEMODE:
			LOG(G3D,"DL Shade mode: %06x", data);
			break;

		case GE_CMD_PATCHDIVISION:
			LOG(G3D,"DL Patch Division: %06x", data);
			break;

		case GE_CMD_MATERIALUPDATE:
			LOG(G3D,"DL Material Update: %d", data);
			break;


		//////////////////////////////////////////////////////////////////
		//	CLEARING
		//////////////////////////////////////////////////////////////////
		case GE_CMD_CLEARMODE:
			if (data & 1)
			{
				glDepthMask((data>>10) & 1); // Update Z?
				glDepthMask(0);
				bool colMask = (data>>8) & 1;
				bool alphaMask = (data>>9) & 1;
				glColorMask(colMask,colMask,colMask,alphaMask);
				//glDisable(GL_BLEND);
			}
			else
			{
				glDepthMask(1);
				glColorMask(1,1,1,1);
			}
			LOG(G3D,"DL Clear mode: %06x", data);
			break;



		//////////////////////////////////////////////////////////////////
		//	ALPHA BLENDING
		//////////////////////////////////////////////////////////////////
		case GE_CMD_ALPHABLENDENABLE:
			LOG(G3D,"DL Alpha blend enable: %d", data);
			glEnDis(GL_BLEND, data);
			break;

		case GE_CMD_BLENDMODE:
			LOG(G3D,"DL Blend mode: %06x", data);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			//glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);
			break;

		case GE_CMD_BLENDFIXEDA:
			LOG(G3D,"DL Blend fix A: %06x", data);
			break;

		case GE_CMD_BLENDFIXEDB:
			LOG(G3D,"DL Blend fix B: %06x", data);
			break;

		case GE_CMD_ALPHATESTENABLE:
			LOG(G3D,"DL Alpha test enable: %d", data);
			glEnDis(GL_ALPHA_TEST, data);
			glAlphaFunc(GL_GREATER,0.03f);			
			break;

		case GE_CMD_TEXFUNC:
			{
				LOG(G3D,"DL TexFunc %i", data&7);
				
				int m=GL_MODULATE;
				switch (data & 7)
				{
				case 0: m=GL_MODULATE; break;
				case 1: m=GL_DECAL; break;
				case 2: m=GL_BLEND; break;
				case 3: m=GL_REPLACE; break;
				case 4: m=GL_ADD; break;
				}

				/*
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB,      GL_MODULATE);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE0_RGB,      GL_CONSTANT);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND0_RGB,     GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_SOURCE1_RGB,      GL_TEXTURE);
				glTexEnvi(GL_TEXTURE_ENV, GL_OPERAND1_RGB,     GL_SRC_COLOR);
				glTexEnvi(GL_TEXTURE_ENV, GL_RGB_SCALE, 1);

				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, m);
				glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_ALPHA, GL_MODULATE);*/
				break;
			}

		//////////////////////////////////////////////////////////////////
		//	Z/STENCIL TESTING
		//////////////////////////////////////////////////////////////////

		case GE_CMD_ZTESTENABLE:
			glEnDis(GL_DEPTH_TEST, data);
			LOG(G3D,"DL Z test enable: %d", data);
			break;

		case GE_CMD_STENCILTESTENABLE:
			LOG(G3D,"DL Stencil test enable: %d", data);
			break;

		case GE_CMD_ZTEST:
			{
				static const GLuint ztests[8] = 
				{
					GL_NEVER, GL_ALWAYS, GL_EQUAL, GL_NOTEQUAL, 
					GL_LESS, GL_LEQUAL, GL_GREATER, GL_GEQUAL
				};
				//glDepthFunc(ztests[data&7]);
				glDepthFunc(GL_LEQUAL);
				LOG(G3D,"DL Z test mode: %i", data);
			}
			break;

		case GE_CMD_MORPHWEIGHT0:
		case GE_CMD_MORPHWEIGHT1:
		case GE_CMD_MORPHWEIGHT2:
		case GE_CMD_MORPHWEIGHT3:
		case GE_CMD_MORPHWEIGHT4:
		case GE_CMD_MORPHWEIGHT5:
		case GE_CMD_MORPHWEIGHT6:
		case GE_CMD_MORPHWEIGHT7:
			gstate.morphWeights[(op>>24)-GE_CMD_MORPHWEIGHT0] = getFloat24(data);
			break;
 
		case GE_CMD_DITH0:
		case GE_CMD_DITH1:
		case GE_CMD_DITH2:
		case GE_CMD_DITH3:
			LOG(G3D,"DL DitherMatrix %i = %06x",(op>>24)-GE_CMD_DITH0,data);
			break;

		case GE_CMD_WORLDMATRIXNUMBER:
			LOG(G3D,"DL World matrix # %i", data);
			gstate.worldmtxnum = data&0xF;
			break;

		case GE_CMD_WORLDMATRIXDATA:
			LOG(G3D,"DL World matrix data # %f", getFloat24(data));
			worldMatrix[gstate.worldmtxnum++] = getFloat24(data);
			break;

		case GE_CMD_VIEWMATRIXNUMBER:
			LOG(G3D,"DL VIEW matrix # %i", data);
			gstate.viewmtxnum = data&0xF;
			break;

		case GE_CMD_VIEWMATRIXDATA:
			LOG(G3D,"DL VIEW matrix data # %f", getFloat24(data));
			viewMatrix[gstate.viewmtxnum++] = getFloat24(data);
			break;

		case GE_CMD_PROJMATRIXNUMBER:
			LOG(G3D,"DL PROJECTION matrix # %i", data);
			gstate.projmtxnum = data&0xF;
			break;

		case GE_CMD_PROJMATRIXDATA:
			LOG(G3D,"DL PROJECTION matrix data # %f", getFloat24(data));
			projMatrix[gstate.projmtxnum++] = getFloat24(data);
			break;

		case GE_CMD_TGENMATRIXNUMBER:
			LOG(G3D,"DL TGEN matrix # %i", data);
			gstate.texmtxnum = data&0xF;
			break;

		case GE_CMD_TGENMATRIXDATA:
			LOG(G3D,"DL TGEN matrix data # %f", getFloat24(data));
			tgenMatrix[gstate.texmtxnum++] = getFloat24(data);
			break;

		case GE_CMD_BONEMATRIXNUMBER:
			LOG(G3D,"DL BONE matrix #%i", data);
			gstate.boneMatrixNumber = data;
			break;

		case GE_CMD_BONEMATRIXDATA:
			LOG(G3D,"DL BONE matrix data #%i %f", gstate.boneMatrixNumber, getFloat24(data));
			boneMatrix[gstate.boneMatrixNumber++] = getFloat24(data);
			break;

		default:
			LOG(G3D,"DL Unknown: %08x @ %08x", op, pc);
			break;

			//ETC...
		}
		pc+=4;
	}
}
